package defaultClient

import (
	"bytes"
	"encoding/json"
	"fmt"
	"gep-demo-go/com/baofoo/api/exception"
	"gep-demo-go/com/baofoo/api/request"
	"gep-demo-go/com/baofoo/api/response"
	"gep-demo-go/com/baofoo/api/utils"
	"io"
	"log"
	"mime/multipart"
	"net/http"
	"net/url"
	"time"
)

const httpTimeout = 30 * time.Second

type DefaultBFClient struct {
	GatewayUrl    string
	PfxPath       string
	PfxPsw        string
	CerPath       string
	DataType      string
	Version       string
	UserNo        string
	CertificateId string
}

func NewDefaultBFClient() *DefaultBFClient {
	return &DefaultBFClient{
		GatewayUrl:    utils.HOST,
		PfxPath:       utils.PFX_PATH, // 商户私钥证书路径
		PfxPsw:        utils.PFX_PSW,  // 证书密码
		CerPath:       utils.CER_PATH, // 宝付公钥证书路径
		DataType:      utils.DATA_TYPE,
		Version:       utils.VERSION,
		UserNo:        utils.USER_NO,
		CertificateId: utils.CERTIFICATE_ID,
	}
}

func (c *DefaultBFClient) Execute(bfRequest *request.BFRequest) (any, error) {
	param, err := json.Marshal(bfRequest.Param)
	if err != nil {
		return nil, &exception.BFLibraryError{Message: "json.Marshal param is fail " + err.Error()}
	}
	log.Println("dataContent 明文:", string(param)) // 打印请求参数
	path := bfRequest.Path
	if path == "" {
		return nil, &exception.BFLibraryError{Message: "path cannot be empty"}
	}

	dataContent, err := utils.EncryptWithPrivateKey(string(param), c.PfxPath, c.PfxPsw) // 加密
	if err != nil {
		return nil, &exception.BFLibraryError{Message: "EncryptWithPrivateKey is fail " + err.Error()}
	}
	log.Println("dataContent 密文:", dataContent) // 打印加密后的数据

	postFields := map[string]string{
		"dataContent":   dataContent,
		"dataType":      c.DataType,
		"version":       c.Version,
		"userNo":        c.UserNo,
		"certificateId": c.CertificateId,
	}

	// 发送 POST 请求
	resp, err := c.DoPost(c.GatewayUrl+path, postFields, bfRequest.FileItems, bfRequest.Headers, bfRequest.HasDownload)

	if err != nil { // 处理错误
		return nil, &exception.BFLibraryError{Message: "DoPost is fail " + err.Error()}
	}
	if bfRequest.HasDownload { // 下载文件,无需解密
		return resp, nil
	}
	var bfResponse response.BfResponse

	respBytes, ok := resp.([]byte)
	if !ok {
		return nil, &exception.BFLibraryError{Message: "response is not of type []byte"}
	}

	err = json.Unmarshal(respBytes, &bfResponse)
	if err != nil {
		return nil, &exception.BFLibraryError{Message: "BFResponse Unmarshal is fail " + err.Error()}
	}
	log.Println("bfResponse:", bfResponse)
	if bfResponse.Success {
		// 解密 result
		result, err := utils.DecryptWithPublicKey(bfResponse.Result, c.CerPath)
		if err != nil {
			return nil, &exception.BFLibraryError{Message: "DecryptWithPublicKey is fail " + err.Error()}
		}
		bfResponse.Result = result
	}
	return bfResponse, nil
}

func (c *DefaultBFClient) DoPost(reqUrl string, param map[string]string, fileItems map[string]utils.FileItem, headers map[string]string, hasDownload bool) (any, error) {
	var b bytes.Buffer
	var writer *multipart.Writer
	postMultipart := len(fileItems) > 0
	if postMultipart {
		writer = multipart.NewWriter(&b)

		// 添加表单字段
		for key, val := range param {
			err := writer.WriteField(key, val)
			if err != nil {
				return nil, &exception.BFLibraryError{Message: "error writing form field: " + err.Error()}
			}
		}

		// 添加文件字段
		for key, fileItem := range fileItems {
			fileName, err := fileItem.GetFileName()
			if err != nil {
				return nil, &exception.BFLibraryError{Message: "error getting file name: " + err.Error()}
			}
			log.Printf("File name for key %s: %s", key, fileName)

			part, err := writer.CreateFormFile(key, fileName)
			if err != nil {
				return nil, &exception.BFLibraryError{Message: "error creating form file: " + err.Error()}
			}

			err = fileItem.WriteFileContent(part)
			if err != nil {
				return nil, &exception.BFLibraryError{Message: "error writing file content: " + err.Error()}
			}
			log.Printf("Successfully wrote file content for key %s with file name %s", key, fileName)

		}

		// 关闭 writer 以确保写入结束
		err := writer.Close()
		if err != nil {
			return nil, &exception.BFLibraryError{Message: "error closing writer: " + err.Error()}
		}
	} else {
		formData := url.Values{}
		for key, val := range param {
			formData.Set(key, val)
		}
		b = *bytes.NewBufferString(formData.Encode())
	}

	// 创建新的请求
	req, err := http.NewRequest("POST", reqUrl, &b)
	if err != nil {
		return nil, &exception.BFLibraryError{Message: "error creating request: " + err.Error()}
	}

	// 设置请求头
	// 如果需要返回的错误信息为英文可以设置Accept-Language: en-US
	if postMultipart {
		req.Header.Set("Content-Type", writer.FormDataContentType())
	} else {
		req.Header.Set("Content-Type", "application/x-www-form-urlencoded;charset=UTF-8")
	}
	for key, value := range headers {
		req.Header.Set(key, value)
	}

	// 创建客户端并设置超时
	client := &http.Client{
		Timeout: httpTimeout,
	}

	// 发送请求
	resp, err := client.Do(req)
	if err != nil {
		return nil, &exception.BFLibraryError{Message: "error sending request: " + err.Error()}
	}
	defer resp.Body.Close()
	if resp.StatusCode != http.StatusOK {
		return nil, &exception.BFLibraryError{Message: fmt.Sprintf("unexpected status code: %d", resp.StatusCode)}
	}

	for key, values := range resp.Header {
		for _, value := range values {
			log.Printf("Header: %s: %s", key, value)
		}
	}

	// 读取响应
	body, err := io.ReadAll(resp.Body)
	if err != nil {
		return nil, &exception.BFLibraryError{Message: "error reading response body: " + err.Error()}
	}
	log.Println("response body:", string(body))
	if hasDownload {
		succ := resp.Header.Get("Success")
		if succ != "true" {
			return nil, &exception.BFLibraryError{Message: "download failed: " + string(body)}
		}
	}
	return body, nil
}
