package utils

import (
	"crypto/md5"
	"encoding/hex"
	"errors"
	"io"
	"mime"
	"os"
	"path/filepath"
)

type FileItem struct {
	fileName string
	mimeType string
	fileMd5  string
	content  []byte
	file     *os.File
}

func NewFileItemFromFile(file *os.File) *FileItem {
	return &FileItem{file: file}
}

func NewFileItemFromFilePath(filePath string) (*FileItem, error) {
	file, err := os.Open(filePath)
	if err != nil {
		return nil, err
	}
	return NewFileItemFromFile(file), nil
}

func NewFileItemFromBytes(fileName string, content []byte) *FileItem {
	return &FileItem{fileName: fileName, content: content}
}

func NewFileItemFromBytesWithMime(fileName string, content []byte, mimeType string) *FileItem {
	return &FileItem{fileName: fileName, content: content, mimeType: mimeType}
}

// GetFileName returns the file name of the FileItem.
func (fi *FileItem) GetFileName() (string, error) {
	if fi.fileName == "" && fi.file != nil {
		fileInfo, err := fi.file.Stat()
		if err != nil {
			return "", err
		}
		fi.fileName = fileInfo.Name()
	}
	return fi.fileName, nil
}

// GetMimeType returns the MIME type of the FileItem.
func (fi *FileItem) GetMimeType() (string, error) {
	if fi.mimeType == "" {
		header, err := fi.getFileHeader()
		if err != nil {
			return "", err
		}
		fi.mimeType = mime.TypeByExtension(filepath.Ext(header))
	}
	return fi.mimeType, nil
}

// getFileHeader reads the first few bytes of the file content.
func (fi *FileItem) getFileHeader() (string, error) {
	if fi.content == nil && fi.file != nil {
		buffer := make([]byte, 16)
		_, err := fi.file.Read(buffer)
		if err != nil {
			return "", err
		}
		return string(buffer), nil
	}
	return string(fi.content), nil
}

// GetContent returns the content of the FileItem as a byte array.
func (fi *FileItem) GetContent() ([]byte, error) {
	if fi.content == nil && fi.file != nil {
		content, err := io.ReadAll(fi.file)
		if err != nil {
			return nil, err
		}
		fi.content = content
	}
	return fi.content, nil
}

// WriteFileContent writes the file content to an output stream.
func (fi *FileItem) WriteFileContent(out io.Writer) error {
	if out == nil {
		return errors.New("output stream is nil")
	}
	fi.file.Seek(0, io.SeekStart) // 重置文件指针到开头,请根据实际场景使用
	if fi.content != nil {
		_, err := out.Write(fi.content)
		return err
	}
	if fi.file != nil {
		defer fi.file.Close() // 确保文件关闭
		_, err := io.Copy(out, fi.file)
		return err
	}

	return errors.New("file content or file is empty")
}

// FileContentLength returns the length of the file content.
func (fi *FileItem) FileContentLength() int64 {
	if fi.content != nil {
		return int64(len(fi.content))
	}
	if fi.file != nil {
		fileInfo, err := fi.file.Stat()
		if err != nil {
			return 0
		}
		return fileInfo.Size()
	}
	return 0
}

// GetFileMd5Encryption returns the MD5 hash of the file content.
func (fi *FileItem) GetFileMd5Encryption() (string, error) {
	if fi.fileMd5 == "" && fi.file != nil {
		hash := md5.New()
		_, err := io.Copy(hash, fi.file)
		if err != nil {
			return "", err
		}
		fi.fileMd5 = hex.EncodeToString(hash.Sum(nil))
	}
	return fi.fileMd5, nil
}

func (fi *FileItem) GetFile() *os.File {
	return fi.file
}
