package utils

import (
	"bytes"
	"crypto"
	"crypto/rand"
	"crypto/rsa"
	"crypto/sha1"
	"crypto/x509"
	"encoding/base64"
	"encoding/hex"
	"encoding/pem"
	"errors"
	"io"
	"math/big"
	"os"

	"golang.org/x/crypto/pkcs12"
)

var (
	ErrDataToLarge     = errors.New("message too long for RSA public key size")
	ErrDataLen         = errors.New("data length error")
	ErrDataBroken      = errors.New("data broken, first byte is not zero")
	ErrKeyPairDismatch = errors.New("data is not encrypted by the private key")
	ErrDecryption      = errors.New("decryption error")
	ErrPublicKey       = errors.New("get public key error")
	ErrPrivateKey      = errors.New("get private key error")
)

func LoadPrivateKey(privateKeyName, privatePassword string) (*rsa.PrivateKey, error) {
	bytes, err := readFile(privateKeyName)
	if err != nil {
		return nil, err
	}

	blocks, err := pkcs12.ToPEM(bytes, privatePassword)
	if err != nil {
		return nil, err
	}
	if len(blocks) != 2 {
		return nil, errors.New("failed to decode PEM block containing private key")
	}
	privateKey, err := x509.ParsePKCS1PrivateKey(blocks[0].Bytes)
	if err != nil {
		return nil, err
	}

	return privateKey, nil
}

func LoadPublicKey(publicName string) (*rsa.PublicKey, error) {
	bytes, err := readFile(publicName)
	if err != nil {
		return nil, err
	}

	block, _ := pem.Decode(bytes)
	if block == nil {
		return nil, errors.New("failed to decode PEM block containing public key")
	}
	x509Cer, err := x509.ParseCertificate(block.Bytes)
	if err != nil {
		return nil, err
	}

	publicKey, ok := x509Cer.PublicKey.(*rsa.PublicKey)
	if !ok {
		return nil, errors.New("failed to assert type *rsa.PublicKey")
	}

	return publicKey, nil
}

func EncryptWithPrivateKey(data, pfxpath, pfxpass string) (string, error) {
	privateKey, err := LoadPrivateKey(pfxpath, pfxpass)
	if err != nil {
		return "", err
	}
	baseData := base64.StdEncoding.EncodeToString([]byte(data))
	return encryptDecryptWithPrivateKey(privateKey, baseData, true)
}

func DecryptWithPublicKey(data, pubpath string) (string, error) {
	dataByte, err := hex.DecodeString(data)
	if err != nil {
		return "", err
	}
	publicKey, err := LoadPublicKey(pubpath)
	if err != nil {
		return "", err
	}
	return encryptDecryptWithPublicKey(publicKey, dataByte, false)
}

func EncryptWithPublicKey(data, cerpath string) (string, error) {
	publicKey, err := LoadPublicKey(cerpath)
	if err != nil {
		return "", err
	}
	baseData := base64.StdEncoding.EncodeToString([]byte(data))
	return encryptDecryptWithPublicKey(publicKey, []byte(baseData), true)
}

func DecryptWithPrivateKey(data, pripath, pfxpass string) (string, error) {
	dataByte, err := hex.DecodeString(data)
	if err != nil {
		return "", err
	}
	privateKey, err := LoadPrivateKey(pripath, pfxpass)
	if err != nil {
		return "", err
	}
	return encryptDecryptWithPrivateKey(privateKey, string(dataByte), false)
}

func SignSha1WithRsa(data, pripath, pfxpass string) (string, error) {
	privateKey, err := LoadPrivateKey(pripath, pfxpass)
	if err != nil {
		return "", err
	}

	sha1Hash := sha1.New()
	sha1Hash.Write([]byte(data))
	hashed := sha1Hash.Sum(nil)
	signByte, err := rsa.SignPKCS1v15(rand.Reader, privateKey, crypto.SHA1, hashed)

	return hex.EncodeToString(signByte), err
}

func VerifySignSha1WithRsa(data, signData, cerpath string) error {
	publicKey, err := LoadPublicKey(cerpath)
	if err != nil {
		return err
	}

	sign, err := hex.DecodeString(signData)
	if err != nil {
		return err
	}
	hash := sha1.New()
	hash.Write([]byte(data))
	return rsa.VerifyPKCS1v15(publicKey, crypto.SHA1, hash.Sum(nil), sign)
}

func pubKeyIO(pub *rsa.PublicKey, in io.Reader, out io.Writer, isEncrypt bool) error {
	k := (pub.N.BitLen() + 7) / 8
	if isEncrypt {
		k -= 11
	}
	buf := make([]byte, k)
	var b []byte
	for {
		size, err := in.Read(buf)
		if err != nil {
			if err == io.EOF {
				return nil
			}
			return err
		}
		if size < k {
			b = buf[:size]
		} else {
			b = buf
		}
		if isEncrypt {
			b, err = rsa.EncryptPKCS1v15(rand.Reader, pub, b)
		} else {
			b, err = pubKeyDecrypt(pub, b)
		}
		if err != nil {
			return err
		}
		if _, err = out.Write(b); err != nil {
			return err
		}
	}
}

func priKeyIO(pri *rsa.PrivateKey, in io.Reader, out io.Writer, isEncrypt bool) error {
	k := (pri.N.BitLen() + 7) / 8
	if isEncrypt {
		k -= 11
	}
	buf := make([]byte, k)
	var b []byte
	for {
		size, err := in.Read(buf)
		if err != nil {
			if err == io.EOF {
				return nil
			}
			return err
		}
		if size < k {
			b = buf[:size]
		} else {
			b = buf
		}
		if isEncrypt {
			b, err = priKeyEncrypt(rand.Reader, pri, b)
		} else {
			b, err = rsa.DecryptPKCS1v15(rand.Reader, pri, b)
		}
		if err != nil {
			return err
		}
		if _, err = out.Write(b); err != nil {
			return err
		}
	}
}

func pubKeyDecrypt(pub *rsa.PublicKey, data []byte) ([]byte, error) {
	k := (pub.N.BitLen() + 7) / 8
	if k != len(data) {
		return nil, ErrDataLen
	}
	m := new(big.Int).SetBytes(data)
	if m.Cmp(pub.N) > 0 {
		return nil, ErrDataToLarge
	}
	m.Exp(m, big.NewInt(int64(pub.E)), pub.N)
	d := leftPad(m.Bytes(), k)
	if d[0] != 0 {
		return nil, ErrDataBroken
	}
	if d[1] != 0 && d[1] != 1 {
		return nil, ErrKeyPairDismatch
	}
	var i = 2
	for ; i < len(d); i++ {
		if d[i] == 0 {
			break
		}
	}
	i++
	if i == len(d) {
		return nil, nil
	}
	return d[i:], nil
}

func priKeyEncrypt(rand io.Reader, priv *rsa.PrivateKey, hashed []byte) ([]byte, error) {
	tLen := len(hashed)
	k := (priv.N.BitLen() + 7) / 8
	if k < tLen+11 {
		return nil, ErrDataLen
	}
	em := make([]byte, k)
	em[1] = 1
	for i := 2; i < k-tLen-1; i++ {
		em[i] = 0xff
	}
	copy(em[k-tLen:k], hashed)
	m := new(big.Int).SetBytes(em)
	c, err := decrypt(rand, priv, m)
	if err != nil {
		return nil, err
	}
	copyWithLeftPad(em, c.Bytes())
	return em, nil
}

func encryptDecryptWithPrivateKey(privateKey *rsa.PrivateKey, data string, isEncrypt bool) (string, error) {
	out := bytes.NewBuffer(nil)
	in := bytes.NewReader([]byte(data))
	if err := priKeyIO(privateKey, in, out, isEncrypt); err != nil {
		return "", err
	}

	result, err := io.ReadAll(out)
	if err != nil {
		return "", err
	}

	if isEncrypt {
		return hex.EncodeToString(result), nil
	}
	dataBs, err := base64.StdEncoding.DecodeString(string(result))
	if err != nil {
		return "", err
	}
	return string(dataBs), nil
}

func encryptDecryptWithPublicKey(publicKey *rsa.PublicKey, data []byte, isEncrypt bool) (string, error) {
	out := bytes.NewBuffer(nil)
	in := bytes.NewReader(data)
	if err := pubKeyIO(publicKey, in, out, isEncrypt); err != nil {
		return "", err
	}

	result, err := io.ReadAll(out)
	if err != nil {
		return "", err
	}

	if isEncrypt {
		return hex.EncodeToString(result), nil
	}
	dataBs, err := base64.StdEncoding.DecodeString(string(result))
	if err != nil {
		return "", err
	}
	return string(dataBs), nil
}

func readFile(fileName string) ([]byte, error) {
	return os.ReadFile(fileName)
}

func leftPad(input []byte, length int) []byte {
	if len(input) >= length {
		return input
	}
	padding := make([]byte, length-len(input))
	return append(padding, input...)
}

func copyWithLeftPad(dst, src []byte) {
	copy(dst[len(dst)-len(src):], src)
}

func decrypt(random io.Reader, priv *rsa.PrivateKey, c *big.Int) (*big.Int, error) {
	if c.Cmp(priv.N) > 0 {
		return nil, ErrDecryption
	}
	var ir *big.Int
	if priv.Precomputed.Dp == nil {
		ir = new(big.Int).Exp(c, priv.D, priv.N)
	} else {
		m := new(big.Int).Exp(c, priv.Precomputed.Dp, priv.Primes[0])
		m2 := new(big.Int).Exp(c, priv.Precomputed.Dq, priv.Primes[1])
		m.Sub(m, m2)
		if m.Sign() < 0 {
			m.Add(m, priv.Primes[0])
		}
		m.Mul(m, priv.Precomputed.Qinv)
		m.Mod(m, priv.Primes[0])
		m.Mul(m, priv.Primes[1])
		m.Add(m, m2)
		ir = m
	}
	return ir, nil
}
